import { Request, Response } from "express";
import logger from "../logger/logger";
import util from "util";
import { exec } from "child_process";
import fs from "fs-extra";
import Website from "../models/website";
import Store from "../models/store";

const execPromise = util.promisify(exec);

// this is a webhook that will be called by github when a new commit is pushed to the website template repository
export async function UpdateTemplateWebhook(req: Request, res: Response) {
  try {
    await cloneWebsiteTemplate();

    res.status(200).send({ message: "ok" });
  } catch (error) {
    console.log("error", error);
    res.status(500).send({ error: "invalid request" });
  }
}

export async function cloneWebsiteTemplate() {
  let tmpDirWebsiteTemplate = process.env
    .WEBSITE_BUILDER_TMP_DIR_WEBSITE_TEMPLATE as string;

  // if website folder exists, pull, else clone

  if (await fs.pathExists(tmpDirWebsiteTemplate)) {
    let { stdout } = await execPromise(`git pull`, {
      cwd: tmpDirWebsiteTemplate,
    });

    console.debug(stdout);
  } else {
    let { stdout } = await execPromise(
      `git clone ${process.env.WEBSITE_BUILDER_TEMPLATE_REPOSITORY_URL} ${tmpDirWebsiteTemplate}`
    );

    console.debug(stdout);
  }

  // execute npm install

  let { stdout: installOutput } = await execPromise(`npm install`, {
    cwd: tmpDirWebsiteTemplate,
  });

  console.debug(installOutput);
  console.debug("website template cloned");
}

export async function CreateWebsite(req: Request, res: Response) {
  try {
    let { storeId } = req.body;

    // validate request

    if (!storeId) {
      res.status(400).send({ error: "invalid request" });
      return;
    }

    // check if website already exists

    const websiteExists = await Website.findOne({
      where: {
        store_id: storeId,
      },
    });

    if (websiteExists) {
      res.status(409).send({ error: "website already exists" });
      return;
    }

    // get store name

    const store = await Store.findOne({
      where: {
        store_id: storeId,
      },
      attributes: ["name"],
    });

    if (!store) {
      res.status(404).send({ error: "store not found" });
      return;
    }

    const storeName = store.name.toLowerCase().replace(/ /g, "-");

    const tmpDirWebsiteTemplate = process.env
      .WEBSITE_BUILDER_TMP_DIR_WEBSITE_TEMPLATE as string;
    const tmpDirCustomerWebsite = `${process.env.WEBSITE_BUILDER_TMP_DIR}/${storeName}`;
    const customerWebsiteDir = `${process.env.WEBSITE_BUILDER_TMP_CUSTOMER_WEBSITES_DIR}/${storeName}`;

    // check if website template folder exists

    if (!(await fs.pathExists(tmpDirWebsiteTemplate))) {
      logger.error("Website template folder does not exist. Clone it first.");
      res.status(500).send({ error: "invalid request" });
      return;
    }

    // copy website-template folder to customer website folder
    await fs.copy(tmpDirWebsiteTemplate, tmpDirCustomerWebsite);

    // TODO: add config file

    // run npm build

    let { stdout: buildOutput } = await execPromise(`npm run build`, {
      cwd: tmpDirCustomerWebsite,
    });

    console.debug(buildOutput);

    // copy the build folder to customer websites folder

    await fs.copy(`${tmpDirCustomerWebsite}/build`, customerWebsiteDir);

    // remove tmp folder

    await fs.remove(tmpDirCustomerWebsite);

    // create database record

    await Website.create({
      website_id: storeName,
      store_id: storeId,
    });

    res.status(200).send({ message: "oks" });
  } catch (error) {
    console.log("error", error);
    res.status(500).send({ error: "invalid request" });
  }
}

export async function GetWebsite(req: Request, res: Response) {
  try {
    const { storeId } = req.params;

    const website = await Website.findOne({
      where: {
        store_id: storeId,
      },
    });

    if (!website) {
      res.status(404).send({ error: "website not found" });
      return;
    }

    res.status(200).json({ message: "ok" });
  } catch (error) {
    console.log("error", error);
    res.status(500).send({ error: "invalid request" });
  }
}
